package com.trolltech.tools.ant;

import org.apache.tools.ant.*;

import java.io.*;
import java.util.List;
import java.util.Map;


class Exec {

    /**
     * Executes the command specified by cmd and returns the printed output
     * from the process' stdout and stderr in the array on position 0 and 1 respectivly.
     * @param cmd The command to execute
     * @return An array of length 2, containing the [stdout, stderr] output.
     * @throws IOException If an error occurs
     * @throws InterruptedException If an error occurs...
     */
    public static String[] execute(String ... cmd) throws IOException, InterruptedException {
        Process p = Runtime.getRuntime().exec(cmd);

        ByteArrayOutputStream outdata = new ByteArrayOutputStream();
        PrintStream out = new PrintStream(outdata);

        ByteArrayOutputStream errdata = new ByteArrayOutputStream();
        PrintStream err = new PrintStream(errdata);

        StreamConsumer stdoutReader = new StreamConsumer(p.getInputStream(), out);
        StreamConsumer stderrReader = new StreamConsumer(p.getErrorStream(), err);
        stdoutReader.start();
        stderrReader.start();
        p.waitFor();

        stdoutReader.join();
        stderrReader.join();
        out.close();
        err.close();

        return new String[] { outdata.toString(), errdata.toString() };
    }

    /**
     * Convenience method for exec(String, File).
     * @param command Command to be executed.
     */
    public static void exec(String command) {
        exec(command, null);
    }

    /**
     * Execute command command in directory dir.
     * @param command Command to be executed.
     * @param dir Directory where command should be executed.
     * @throws BuildException Thrown if process exit value is not zero or IOException has been occurred.
     */
    public static void exec(String command, File dir) throws BuildException {
        String directory = ((dir != null) ? "(" + Util.makeCanonical(dir) + ")" : "");
        System.out.println("Running : " + directory + " " + command);
        try {
            Process process = Runtime.getRuntime().exec(command, null, dir);
            Util.redirectOutput(process);
            if (process.exitValue() != 0) {
                throw new BuildException("Running: " + command + " failed with exit code: " + process.exitValue());
            }
        } catch (IOException e) {
            throw new BuildException("Running: " + command + " failed with error message: " + e.getMessage(), e);
        }
    }

    /**
     * TODO: this should be merged with above one, repeating code is not that wise.
     * 
     * Executes process in more verbose manner.
     * @param cmd Array of command and its arguments to be executed.
     * @param dir Directory where should be executed.
     * @param verbose Whether to be verbose.
     * @throws BuildException Thrown if process exit value is not zero or IOException has been occurred.
     */
    public static void exec(String cmd[], File dir, boolean verbose) throws BuildException {
        if (verbose) {
            StringBuilder b = new StringBuilder();
            for (String s : cmd)
                b.append(s).append(' ');
            System.out.println("Running : " + ((dir!=null)? "(" + Util.makeCanonical(dir) + ")" : "") + " " + b);
        }

        try {
            Process process = Runtime.getRuntime().exec(cmd, null, dir);
            Util.redirectOutput(process);
            if (process.exitValue() != 0) {
                throw new BuildException("Running: '" + join(cmd) + "' failed.");
            }
        } catch (IOException e) {
            throw new BuildException("Running: '" + join(cmd) + "' failed.", e);
        }
    }
    
    public static void execute(List<String> command, String ldpath, File directory) throws BuildException {
        ProcessBuilder builder = new ProcessBuilder(command);
        
        //NOTE: this is most likely very linux-specific system. For Windows one would use PATH instead,
        //but it should not be needed there in first place... Only if you want to have same kind of building
        //environment one can have for Linux.
        Map<String, String> env = builder.environment();
        //we don't need other stuff to LD_LIBRARY_PATH
        //String ldpath = env.get("LD_LIBRARY_PATH");
        //System.out.println("ldpath: " + ldpath);
        //ldpath = "" + ldpath;
        env.put("LD_LIBRARY_PATH", ldpath);
        System.out.println("ldpath: " + ldpath);
        builder.directory(directory);
        try {
			Process process = builder.start();
			Util.redirectOutput(process);
            if (process.exitValue() != 0) {
                throw new BuildException("Running: '" + command.toString() + "' failed.");
            }
		} catch (IOException e) {
			 //TODO: this may not work
			 throw new BuildException("Running: '" + command.toString() + "' failed.", e);
		}
    }
    
    /**
     * Internal helper of Exec.
     * @param ar What to join
     * @return array joined together to form "foo1, foo2, .."
     */
    private static String join(String ar[]) {
        String s = "";
        for (int i = 0; i<ar.length; ++i) {
            s += ar[i];
            if (i < ar.length - 1)
                s += ", ";
        }
        return s;
    }

}